@@ -8,10 +8,11 @@ from collections.abc import AsyncIterable
 from contextlib import suppress
 from io import BytesIO
 from types import MappingProxyType, TracebackType
-from typing import Any, Dict, Optional, Tuple, Type, Union
+from typing import (
+    Any, Awaitable, Callable, Dict, Mapping, Optional, Tuple, Type, Union,
+)
 
 import pamqp.frame
-import pkg_resources
 from pamqp import commands as spec
 from pamqp.base import Frame
 from pamqp.common import FieldTable
@@ -30,16 +31,24 @@ from .auth import AuthMechanism
 from .base import Base, task
 from .channel import Channel
 from .exceptions import (
-    AMQPError, AuthenticationError, ConnectionChannelError, ConnectionClosed,
-    ConnectionCommandInvalid, ConnectionFrameError, ConnectionInternalError,
-    ConnectionNotAllowed, ConnectionNotImplemented, ConnectionResourceError,
-    ConnectionSyntaxError, ConnectionUnexpectedFrame, IncompatibleProtocolError,
-    ProbableAuthenticationError,
+    AMQPConnectionError, AMQPError, AuthenticationError, ConnectionChannelError,
+    ConnectionClosed, ConnectionCommandInvalid, ConnectionFrameError,
+    ConnectionInternalError, ConnectionNotAllowed, ConnectionNotImplemented,
+    ConnectionResourceError, ConnectionSyntaxError, ConnectionUnexpectedFrame,
+    IncompatibleProtocolError, ProbableAuthenticationError,
 )
 from .tools import Countdown, censor_url
 
 
-__version__ = pkg_resources.get_distribution("aiormq").version
+# noinspection PyUnresolvedReferences
+try:
+    from importlib.metadata import Distribution
+    __version__ = Distribution.from_name("aiormq").version
+except ImportError:
+    import pkg_resources
+    __version__ = pkg_resources.get_distribution("aiormq").version
+
+
 log = logging.getLogger(__name__)
 
 CHANNEL_CLOSE_RESPONSES = (spec.Channel.Close, spec.Channel.CloseOk)
@@ -143,37 +152,56 @@ class FrameReceiver(AsyncIterable):
             del self.reader
             raise StopAsyncIteration
 
-        async with self.lock:
-            try:
-                frame_header = await self.reader.readexactly(1)
+        with BytesIO() as fp:
+            async with self.lock:
+                try:
+                    fp.write(await self.reader.readexactly(1))
 
-                if frame_header == b"\0x00":
-                    raise AMQPFrameError(
-                        await self.reader.read(),
+                    if fp.getvalue() == b"\0x00":
+                        fp.write(await self.reader.read())
+                        raise AMQPFrameError(fp.getvalue())
+
+                    if self.reader is None:
+                        raise AMQPConnectionError()
+
+                    fp.write(await self.reader.readexactly(6))
+
+                    if not self.started and fp.getvalue().startswith(b"AMQP"):
+                        raise AMQPSyntaxError
+                    else:
+                        self.started = True
+
+                    frame_type, _, frame_length = pamqp.frame.frame_parts(
+                        fp.getvalue(),
                     )
+                    if frame_length is None:
+                        raise AMQPInternalError("No frame length", None)
 
-                if self.reader is None:
-                    raise ConnectionError
+                    fp.write(await self.reader.readexactly(frame_length + 1))
+                except asyncio.IncompleteReadError as e:
+                    raise AMQPConnectionError(
+                        "Server connection unexpectedly closed. "
+                        f"Read {len(e.partial)} bytes but {e.expected} "
+                        "bytes expected",
+                    ) from e
+                except ConnectionRefusedError as e:
+                    raise AMQPConnectionError(
+                        f"Server connection refused: {e!r}",
+                    ) from e
+                except ConnectionResetError as e:
+                    raise AMQPConnectionError(
+                        f"Server connection reset: {e!r}",
+                    ) from e
+                except ConnectionError as e:
+                    raise AMQPConnectionError(
+                        f"Server connection error: {e!r}",
+                    ) from e
+                except OSError as e:
+                    raise AMQPConnectionError(
+                        f"Server communication error: {e!r}",
+                    ) from e
 
-                frame_header += await self.reader.readexactly(6)
-
-                if not self.started and frame_header.startswith(b"AMQP"):
-                    raise AMQPSyntaxError
-                else:
-                    self.started = True
-
-                frame_type, _, frame_length = pamqp.frame.frame_parts(
-                    frame_header,
-                )
-                if frame_length is None:
-                    raise AMQPInternalError("No frame length", None)
-
-                frame_payload = await self.reader.readexactly(frame_length + 1)
-            except asyncio.IncompleteReadError as e:
-                raise AMQPFrameError(
-                    "Server connection unexpectedly closed",
-                ) from e
-        return pamqp.frame.unmarshal(frame_header + frame_payload)
+            return pamqp.frame.unmarshal(fp.getvalue())
 
     async def __anext__(self) -> ReceivedFrame:
         return await self.get_frame()
@@ -269,7 +297,7 @@ class Connection(Base, AbstractConnection):
         self.__update_secret_future: Optional[asyncio.Future] = None
         self.__connection_unblocked: asyncio.Event = asyncio.Event()
         self.__heartbeat_grace_timeout = (
-            (self.timeout + 1) * self.HEARTBEAT_GRACE_MULTIPLIER
+            (self.heartbeat_timeout + 1) * self.HEARTBEAT_GRACE_MULTIPLIER
         )
         self.__last_frame_time: float = self.loop.time()
 
@@ -362,7 +390,9 @@ class Connection(Base, AbstractConnection):
         frame_receiver: FrameReceiver,
         wait_response: bool = True,
     ) -> Optional[FrameTypes]:
+
         writer.write(pamqp.frame.marshal(request, 0))
+        await writer.drain()
 
         if not wait_response:
             return None
@@ -402,7 +432,7 @@ class Connection(Base, AbstractConnection):
 
             frame_receiver = FrameReceiver(reader)
         except OSError as e:
-            raise ConnectionError(*e.args) from e
+            raise AMQPConnectionError(*e.args) from e
 
         frame: Optional[FrameTypes]
 
@@ -459,7 +489,7 @@ class Connection(Base, AbstractConnection):
 
             if not isinstance(frame, spec.Connection.OpenOk):
                 raise AMQPInternalError("Connection.OpenOk", frame)
-        except Exception as e:
+        except BaseException as e:
             await self.__close_writer(writer)
             await self.close(e)
             raise
@@ -471,13 +501,8 @@ class Connection(Base, AbstractConnection):
         # noinspection PyAsyncCall
         self._writer_task = self.create_task(self.__writer(writer))
 
-        # Not very optimal, but avoid creating a task for each frame sending
-        # noinspection PyAsyncCall
-        self.create_task(self.__heartbeat())
-
         self.connection_tune = connection_tune
         self.server_properties = server_properties
-        self.__connection_unblocked.set()
         return True
 
     def _on_reader_done(self, task: asyncio.Task) -> None:
@@ -501,9 +526,84 @@ class Connection(Base, AbstractConnection):
 
         self.loop.create_task(close_writer_task())
 
+    async def __handle_close_ok(self, _: spec.Connection.CloseOk) -> None:
+        return
+
+    async def __handle_heartbeat(self, _: Heartbeat) -> None:
+        return
+
+    async def __handle_close(self, frame: spec.Connection.Close) -> None:
+        log.exception(
+            "Unexpected connection close from remote \"%s\", "
+            "Connection.Close(reply_code=%r, reply_text=%r)",
+            self, frame.reply_code, frame.reply_text,
+        )
+
+        self.write_queue.put_nowait(
+            ChannelFrame.marshall(
+                channel_number=0,
+                frames=[spec.Connection.CloseOk()],
+            ),
+        )
+
+        exception = exception_by_code(frame)
+
+        if (
+            self.__update_secret_future is not None and
+            not self.__update_secret_future.done()
+        ):
+            self.__update_secret_future.set_exception(exception)
+        raise exception
+
+    async def __handle_channel_close_ok(
+        self, _: spec.Channel.CloseOk,
+    ) -> None:
+        self.channels.pop(0, None)
+
+    async def __handle_channel_update_secret_ok(
+        self, frame: spec.Connection.UpdateSecretOk,
+    ) -> None:
+        if (
+            self.__update_secret_future is not None and
+            not self.__update_secret_future.done()
+        ):
+            self.__update_secret_future.set_result(frame)
+            return
+        log.warning("Got unexpected UpdateSecretOk frame")
+
+    async def __handle_connection_blocked(
+        self, frame: spec.Connection.Blocked,
+    ) -> None:
+        log.warning("Connection %r was blocked by: %r", self, frame.reason)
+        self.__connection_unblocked.clear()
+
+    async def __handle_connection_unblocked(
+        self, _: spec.Connection.Unblocked,
+    ) -> None:
+        log.warning("Connection %r was unblocked", self)
+        self.__connection_unblocked.set()
+
     async def __reader(self, frame_receiver: FrameReceiver) -> None:
+        self.__connection_unblocked.set()
         self.connected.set()
 
+        # Not very optimal, but avoid creating a task for each frame sending
+        # noinspection PyAsyncCall
+        self.create_task(self.__heartbeat())
+
+        channel_frame_handlers: Mapping[Any, Callable[[Any], Awaitable[None]]]
+        channel_frame_handlers = {
+            spec.Connection.CloseOk: self.__handle_close_ok,
+            spec.Connection.Close: self.__handle_close,
+            Heartbeat: self.__handle_heartbeat,
+            spec.Channel.CloseOk: self.__handle_channel_close_ok,
+            spec.Connection.UpdateSecretOk: (
+                self.__handle_channel_update_secret_ok
+            ),
+            spec.Connection.Blocked: self.__handle_connection_blocked,
+            spec.Connection.Unblocked: self.__handle_connection_unblocked,
+        }
+
         try:
             async for weight, channel, frame in frame_receiver:
                 self.__last_frame_time = self.loop.time()
@@ -514,58 +614,13 @@ class Connection(Base, AbstractConnection):
                 )
 
                 if channel == 0:
-                    if isinstance(frame, spec.Connection.CloseOk):
-                        return
+                    handler = channel_frame_handlers.get(type(frame))
 
-                    if isinstance(frame, spec.Connection.Close):
-                        log.exception(
-                            "Unexpected connection close from remote \"%s\", "
-                            "Connection.Close(reply_code=%r, reply_text=%r)",
-                            self, frame.reply_code, frame.reply_text,
-                        )
-
-                        self.write_queue.put_nowait(
-                            ChannelFrame(
-                                channel_number=0,
-                                frames=[spec.Connection.CloseOk()],
-                            ),
-                        )
-
-                        exception = exception_by_code(frame)
-
-                        if (
-                            self.__update_secret_future is not None and
-                            not self.__update_secret_future.done()
-                        ):
-                            self.__update_secret_future.set_exception(exception)
-
-                        raise exception
-                    elif isinstance(frame, Heartbeat):
-                        continue
-                    elif isinstance(frame, spec.Channel.CloseOk):
-                        self.channels.pop(channel, None)
-                    elif isinstance(frame, spec.Connection.UpdateSecretOk):
-                        if (
-                            self.__update_secret_future is not None and
-                            not self.__update_secret_future.done()
-                        ):
-                            self.__update_secret_future.set_result(frame)
-                        else:
-                            log.warning("Got unexpected UpdateSecretOk frame")
-                        continue
-                    elif isinstance(frame, spec.Connection.Blocked):
-                        log.warning(
-                            "Connection %r was blocked by: %r",
-                            self, frame.reason,
-                        )
-                        self.__connection_unblocked.clear()
-                        continue
-                    elif isinstance(frame, spec.Connection.Unblocked):
-                        log.warning("Connection %r was unblocked", self)
-                        self.__connection_unblocked.set()
+                    if handler is None:
+                        log.error("Unexpected frame %r", frame)
                         continue
 
-                    log.error("Unexpected frame %r", frame)
+                    await handler(frame)
                     continue
 
                 ch: Optional[AbstractChannel] = self.channels.get(channel)
@@ -579,12 +634,14 @@ class Connection(Base, AbstractConnection):
                     self.channels[channel] = None
 
                 await ch.frames.put((weight, frame))
-        except asyncio.CancelledError as e:
+        except asyncio.CancelledError:
             if self.is_connection_was_stuck:
-                raise asyncio.TimeoutError(
-                    "Server connection was stucked. No frames were "
-                    f"received in {self.__heartbeat_grace_timeout} seconds.",
-                ) from e
+                log.warning(
+                    "Server connection %r was stuck. No frames were received "
+                    "in %d seconds.", self, self.__heartbeat_grace_timeout,
+                )
+                self._reader_task.cancel()
+                self._writer_task.cancel()
             raise
 
     @property
@@ -594,7 +651,9 @@ class Connection(Base, AbstractConnection):
 
     async def __heartbeat(self) -> None:
         heartbeat_timeout = max(1, self.heartbeat_timeout // 2)
-        heartbeat = ChannelFrame(frames=[Heartbeat()], channel_number=0)
+        heartbeat = ChannelFrame.marshall(
+            frames=[Heartbeat()], channel_number=0,
+        )
 
         while not self.closing.done():
             if self.is_connection_was_stuck:
@@ -602,7 +661,15 @@ class Connection(Base, AbstractConnection):
                 return
 
             await asyncio.sleep(heartbeat_timeout)
-            await self.write_queue.put(heartbeat)
+
+            try:
+                await asyncio.wait_for(
+                    self.write_queue.put(heartbeat),
+                    timeout=self.__heartbeat_grace_timeout,
+                )
+            except asyncio.TimeoutError:
+                self._reader_task.cancel()
+                return
 
     async def __writer(self, writer: asyncio.StreamWriter) -> None:
         channel_frame: ChannelFrame
@@ -619,29 +686,18 @@ class Connection(Base, AbstractConnection):
             async for channel_frame in frame_iterator:
                 log.debug("Prepare to send %r", channel_frame)
 
-                frame: FrameTypes
-                with BytesIO() as fp:
-                    for frame in channel_frame.frames:
-                        fp.write(
-                            pamqp.frame.marshal(
-                                frame, channel_frame.channel_number,
-                            ),
-                        )
-
-                        if isinstance(frame, spec.Connection.CloseOk):
-                            writer.write(fp.getvalue())
-                            return
-
-                    writer.write(fp.getvalue())
-
-                if (
-                    channel_frame.drain_future is not None and
-                    not channel_frame.drain_future.done()
-                ):
+                writer.write(channel_frame.payload)
+                if channel_frame.should_close:
                     await writer.drain()
-                    channel_frame.drain_future.set_result(None)
+                    channel_frame.drain()
+                    return
+
+                if channel_frame.should_drain:
+                    await writer.drain()
+                    channel_frame.drain()
+
         except asyncio.CancelledError:
-            if not self.__check_writer(writer):
+            if not self.__check_writer(writer) or self.is_connection_was_stuck:
                 raise
 
             frame = spec.Connection.Close(
@@ -651,14 +707,17 @@ class Connection(Base, AbstractConnection):
                 method_id=self.__close_method_id,
             )
 
-            writer.write(pamqp.frame.marshal(frame, 0))
+            writer.write(ChannelFrame.marshall(0, [frame]).payload)
+
             log.debug("Sending %r to %r", frame, self)
 
-            await asyncio.gather(
-                writer.drain(),
-                self.__close_writer(writer),
-                return_exceptions=True,
-            )
+            try:
+                await asyncio.wait_for(
+                    writer.drain(), timeout=self.__heartbeat_grace_timeout,
+                )
+            finally:
+                await self.__close_writer(writer)
+
             raise
         finally:
             log.debug("Writer exited for %r", self)
@@ -670,10 +729,11 @@ class Connection(Base, AbstractConnection):
     else:
         async def __close_writer(self, writer: asyncio.StreamWriter) -> None:
             log.debug("Writer on connection %s closed", self)
-            if writer.can_write_eof():
-                writer.write_eof()
-            writer.close()
-            await writer.wait_closed()
+            with suppress(OSError, RuntimeError):
+                if writer.can_write_eof():
+                    writer.write_eof()
+                writer.close()
+                await writer.wait_closed()
 
     @staticmethod
     def __check_writer(writer: asyncio.StreamWriter) -> bool:
@@ -785,7 +845,7 @@ class Connection(Base, AbstractConnection):
         self, new_secret: str, *,
         reason: str = "", timeout: TimeoutType = None,
     ) -> spec.Connection.UpdateSecretOk:
-        channel_frame = ChannelFrame(
+        channel_frame = ChannelFrame.marshall(
             channel_number=0,
             frames=[
                 spec.Connection.UpdateSecret(
